
#include <eel.h>
#include "util.h"

/* Fill prefix commands for Epsilon.

   Created 2/15/95 by Stever

   Use:
   M-Q          Fill the current paragraph.  Determine the fill prefix
                by taking the longest of: the substring the current
                line has in common with the next line or with the
                previous line.

                > ABC    with cursor on DEF line, prefix=>>
                >>DEF    with cursor on ABC line, prefix=>
                >>GHI

   C-U M-Q      Fill the current paragraph.  Use from start-of-line to
                current cursor position as the fill prefix.
                > ABC    with cursor on the B, prefix="> A"
                >>DEF

   (c) Copyright 1995, Stephen Robbins.  All rights reserved.

*/

#include <eel.h>

#define MAX_PREF_LEN    132             // Maximum prefix length

char *guess_at_prefix();
char *this_line_prefix();
int on_boundary_line ();

void new_smartfill_fill_paragraph()
{
        char *prefix;
        int prefix_len;
        spot z = alloc_spot();          // ***DEBUG***
        
        spot fill_start = alloc_spot();
        spot fill_end = alloc_spot();
        if (iter == 1)
                prefix = guess_at_prefix(&prefix_len);
        else
                prefix = this_line_prefix(&prefix_len);
        say ("Using prefix: '%s'", prefix);
        if (*prefix != '\0') {
                iter = 1;
                set_fill_range (prefix, prefix_len, fill_start, fill_end);
/*
*z=point; point=*fill_start; stuff("==>"); point=*z;
*z=point; point=*fill_end;   stuff("<=="); point=*z;
refresh();
delete (*fill_start, *fill_start+3);
delete (*fill_end,   *fill_end+3);
refresh();
*/
                remove_prefix (prefix_len, fill_start, fill_end);
//refresh ();
                region_fill (*fill_start, *fill_end);
//refresh();
                restore_prefix (prefix, fill_start, fill_end);
//refresh();
                }
        else
                smartfill_fill_paragraph (); // Call orig fill paragraph

        free (prefix);
        free_spot (fill_start);
        free_spot (fill_end);
        free_spot (z);                  // ***DEBUG***
}

REPLACE_FUNC("smartfill", "fill_paragraph");

/* this_lines_prefix(int *prefix_len)

   Returns: a pointer to a newly MALLOC'd string containing this
            line's fill prefix.  The prefix consists of all
            characters before the cursor on the current line.

            prefix_len  set to the length of the prefix
*/

char *this_line_prefix(prefix_len)
int *prefix_len;
{
        char *c;

        *prefix_len = point - give_begin_line();
        c = malloc (*prefix_len + 1);
        grab (give_begin_line(), point, c);
        return c;
}

/* guess_at_prefix (int *prefix_len)

   Returns: a pointer to a newly MALLOC'd string containing this
            line's fill prefix.  The prefix is the longer of:
              (a) the initial string this line has in common with the 
                  previous buffer line.
              (b) the initial string this line has in common with the
                  next buffer line.

            There's one caveat, which is that the prefix must end in
            a non-alphabetic character.  (Otherwise, two line starting
            with the same letter or the same first couple of letters
            could inadvertantly include the letters as part of the
            prefix.)
            
            prefix_len  set to the length of the prefix
*/

char *guess_at_prefix (prefix_len)
int *prefix_len;
{
        int orig_point = point;         // bufptr to original buffer posn
        int orig_line;                  // bufptr to start of original line
        int pref_len1 = 0;              // prefix length prev line's prefix
        int pref_len2 = 0;              // prefix length next line's prefix
        int saved_len = 0;              // temporary-saved prefix len
        char orig_c;                    // char from original line
        char othr_c;                    // char from prev/next line
        char *c;
        
        to_begin_line();
        orig_line = point;
        
        // Find the length of the common substring w/ previous line.
        if (point > narrow_start) {     // there's a prev line.  Find prefix.
                nl_reverse ();
                to_begin_line ();
                while (orig_line + pref_len1 < size()-narrow_end
                       && (orig_c = character (orig_line + pref_len1)) != '\n'
                       && (othr_c = character (point     + pref_len1)) != '\n'
                       && orig_c == othr_c)  pref_len1++;
                // Back up over alphabetics at the end of the prefix:
                saved_len = pref_len1;
                while (pref_len1 > 0
                       && (isalpha (orig_c = character (orig_line+pref_len1-1))
                           || isdigit (orig_c)
                           || isspace (orig_c)))
                                        pref_len1--;
                // And move forward again over spaces.
                while (pref_len1 < saved_len
                       && isspace (character (orig_line+pref_len1)))
                                pref_len1++;
                point = orig_point;
                }

        // Find the length of the common substring with the next line.
        if (nl_forward () != 0) {       // there's a next line.  Find prefix.
                while (point + pref_len2 < size()-narrow_end
                       && (orig_c = character (orig_line + pref_len2)) != '\n'
                       && (othr_c = character (point     + pref_len2)) != '\n'
                       && orig_c == othr_c)  pref_len2++;
                // Back up over alphabetics at the end of the prefix:
                saved_len = pref_len2;
                while (pref_len2 > 0
                       && (isalpha (orig_c = character (orig_line+pref_len2-1))
                           || isdigit (orig_c)
                           || isspace (orig_c)))
                                        pref_len2--;
                // And move forward again over spaces.
                while (pref_len2 < saved_len
                       && isspace (character (orig_line+pref_len2)))
                                pref_len2++;
                point = orig_point;
                }

        *prefix_len = pref_len1 > pref_len2 ? pref_len1 : pref_len2;
        c = malloc (*prefix_len + 1);
        grab (orig_line, orig_line + *prefix_len, c);
        return c;
}

/* remove_prefix (prefix_len, fill_start, fill_end)

   Pass: fill_start points right before the first character of the
   first line of the fill region.  Fill_end points right after the
   last character of the last line of the fill region.

   Assumptions: Every line in the fill region is at least 'prefix_len'
   characters long.
   
   Effects: Removes the first prefix_len characters from all lines
   from fill_start to fill_end in the buffer.
*/

void remove_prefix (prefix_len, fill_start, fill_end)
int prefix_len;
spot fill_start, fill_end;
{
        if (prefix_len == 0) return;

        save_spot point = *fill_start;
        while (point < *fill_end) {
                delete (point, point + prefix_len);
                nl_forward ();
                }
}

/* restore_prefix (prefix, fill_start, fill_end)

   Pass: fill_start - beginning of first line of fill region.
         fill_end   - end of last line of fill region.

   Effects:  Inserts 'prefix' at the start of each line in the
   given fill region.

*/

void restore_prefix (prefix, fill_start, fill_end)
char *prefix;
spot fill_start, fill_end;
{
        if (*prefix == '\0') return;

        save_spot point = *fill_start;
        while (point < *fill_end) {
                stuff (prefix);
                nl_forward ();
                }

        // when the point started out exactly to the left of the
        // first line's prefix, then the prefix we STUFFed during
        // restore got put to the RIGHT of the point, rather than
        // the left.  So fix it.
        restore_vars();
        if (point == *fill_start) point += strlen(prefix);
        
}

/*  set_fill_range (prefix, prefix_len, fill_start, fill_end)

    Pass:  the fill prefix string and the length of the string.
    Pointers to two spots, which will be left with spots
    corresponding to the start of the fill region and the end
    of the fill region.

    Return:  fill_start - right at the start of the line @ the
                          start of the fill region
             fill_end - before the \n of the region's last line

    The region will extend from the closest boundary line
    above to the closest boundary line below the point,
    EXCLUDING the boundary lines.  A boundary line is a blank
    line or a line containing only the fill prefix and blank
    text.

    (1) If ORIG_LINE starts out at the beginning of the buffer,
    the stuff("\n") will insert AFTER orig_line.  orig_line needs to
    be moved forward so that it's within the boundaries of what
    was originally the buffer.
*/  

void set_fill_range (prefix, prefix_len, fill_start, fill_end)
char *prefix;
int prefix_len;
spot fill_start, fill_end;
{
        spot orig_line = alloc_spot ();
        save_spot point;

        // Surround the buffer with <newline>s, to insure there will
        // always be a boundary line before and after the current paragraph.
        point = narrow_start; stuff ("\n");
        point = size()-narrow_end; stuff ("\n");
//refresh();
        // (1)
        point = ((*orig_line <= narrow_start) ? narrow_start+1 : *orig_line);
//stuff("<>");
//refresh();
//delete (point-2, point);
//refresh();

        to_begin_line();
        *orig_line = point;
        while (!on_boundary_line (prefix, prefix_len)) {
                nl_reverse ();
                to_begin_line ();
                }
        nl_forward ();
        *fill_start = point;

        point = *orig_line;
        while (!on_boundary_line (prefix, prefix_len)) {
                nl_forward ();
                }
        *fill_end = --point;

        // remove the surrounding '\n's:
        delete (narrow_start, narrow_start+1);
        delete (size()-narrow_end-1, size()-narrow_end);
}

/* on_boundary_line (char *prefix, int prefix_len)

   Returns: TRUE iff the point is at the start of a line that is a
   fill prefix boundary.  A fill prefix boundary is:
           a blank line
           a line containing only the fill prefix and whitespace
           a line that doesn't start with the fill prefix
*/

int on_boundary_line (prefix, prefix_len)
char *prefix;
int prefix_len;
{
//stuff("==>");
//refresh ();
//delete(point-3,point);
//refresh();
        if (character(point) == '\n') return TRUE; // blank line
        if (match (prefix, FALSE))      // prefix w/ or w/o blanks
                return all_blanks (point+prefix_len, give_end_line ());
        return TRUE;                    // not the correct prefix
}

//////////////////////////////////////////
// Test cases
//////////////////////////////////////////

/*

(1) one line in buffer with \n
(2) one line in buffer w/o \n
(3) empty buffer
(4) two line buffer, on 1st line, 2nd line
(5) three line buffer, on 1st line, 2nd line, 3rd line
(6) paragraph bounded by blank lines
(7) paragraph bounded by lines w/ only the fill prefix
(8) paragraph bounded by lines w/ only the fill prefix,
        in turn bounded by blank lines
(9) test (* = cursor position):
    ABCDE    ABC*       ABC     ABC
    ABC*     ABCDE      ABC*DE  ABCDE*
(10) test all those things with the buffer narrowed to
exactly the region that will be filled;
(11) narrow to include top boundary
(12) narrow to include bottom boundary
(13) narrow to include both boundaries
(14) narrow to only the one line the cursor is on

*/
